#include <stdio.h>
#include <stdlib.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "ledmatrix7219d88.h"
#include "max7219.h"
#include "text.h"
#include "led_snake.h"

//se folosesc PORTD2-DIN; PORTD3-CLK; PORTD4-LOAD;int game_area[MSIZE][MSIZE];

int head, tail;
int direction;
int snake_size;
int speed;

int text_rolling, direction_changed;

int game_area[MSIZE][MSIZE];

coordinate snake[NLEDS];
coordinate food;

// initialize the snake tiles
void init_snake()
{
	for(int i = 0; i < MSIZE; ++i)
		for(int j = 0; j < MSIZE; ++j) {
			game_area[i][j] = EMPTY;
		}
		
	snake_size = SSIZE;
	direction = DOWN;
	
	for(int i = 0; i < snake_size; ++i) {
		snake[i].x = 3;
		snake[i].y = 3 + i;
		game_area[snake[i].x][snake[i].y] = SNAKE;
	}
	
	head = snake_size - 1;
	tail = 0;
}

int select_speed()
{	

	for(int i = 0; i < NLEDS; ++i)
		setledoff(MX, i);
		
	_delay_ms(200);
	int nleds = 3;
	while(1)
	{
		for(int i = 0; i < 6; ++i)
			setled(MX, 51 - 8 * i, (i < nleds));
		
		if ((PINB & (1 << PB1)) == 0)
			nleds = (nleds == 1) ? 1 : (nleds - 1);
		if ((PINB & (1 << PB4)) == 0)
			nleds = (nleds == 6) ? 6 : (nleds + 1);
		if ((PINB & (1 << PB2)) == 0)
			break;
		_delay_ms(200);
	}
	
	return 700 - (nleds * 100);
}

// change the snake direction based on the pressed button
void change_direction()
{
	if(direction_changed)
		return;
	if (
		(PINB & (1 << PB1)) == 0
		&& direction != RIGHT
		&& direction != LEFT
		)
		direction = LEFT;
	if (
		(PINB & (1 << PB4)) == 0
		&& direction != RIGHT
		&& direction != LEFT
		)
		direction = RIGHT;
	if (
		(PINB & (1 << PB0)) == 0
		&& direction != DOWN
		&& direction != UP
		)
		direction = UP;
	if (
		(PINB & (1 << PB3)) == 0
		&& direction != DOWN
		&& direction != UP
		)
		direction = DOWN;
	direction_changed = 1;
}

// external interrupt handler
ISR(PCINT1_vect)
{
	if (
	(PINB & (1 << PB2)) == 0
	&& text_rolling
	)
		text_rolling = 0;

	change_direction();
}

// spawns snake food on an empty tile
void spawn_food()
{
	do {
		food.x = rand() % MSIZE;
		food.y = rand() % MSIZE;
	} while(game_area[food.x][food.y] != EMPTY);
	
	game_area[food.x][food.y] = FOOD;
}

// print the game area to the led matrix
void print_to_matrix(int x[][MSIZE])
{
	for(int i = 0; i < MSIZE; ++i)
		for(int j = 0; j < MSIZE; ++j) {
			setled(MX, MSIZE * i + j, x[i][j]);
		}
}

// play a sound with the given frequency and duration
void play_sound(float frequency, int duration) {
	float period = 100000 / frequency;
	int i, j;

	for (i = 0; i < duration; ++i) {
	
		PORTB |= (1 << PB6);

		for (j = 0; j < period / 2; ++j)
			_delay_us(1);

		PORTB &= ~(1 << PB6);

		for (j = 0 ; j < period / 2; ++j)
			_delay_us(1);
	}
}

void play_food_sound()
{
	play_sound(3000, 20);
	play_sound(5000, 20);
}

void play_death_sound()
{
	play_sound(300, 20);
	play_sound(210, 30);
}

// move the snake one tile
int move()
{
	int previous_head = head;
	
	head = SAFE_INC(head, NLEDS);
	
	switch (direction) {
	case LEFT:
		snake[head].x = SAFE_INC(snake[previous_head].x, MSIZE);
		snake[head].y = snake[previous_head].y;
		break;
	case RIGHT:
		snake[head].x = SAFE_DEC(snake[previous_head].x, MSIZE);
		snake[head].y = snake[previous_head].y;
		break;
	case UP:
		snake[head].x = snake[previous_head].x;
		snake[head].y = SAFE_DEC(snake[previous_head].y, MSIZE);
		break;
	case DOWN:
		snake[head].x = snake[previous_head].x;
		snake[head].y = SAFE_INC(snake[previous_head].y, MSIZE);
	}
	
	int next_tile = game_area[snake[head].x][snake[head].y];
	
	// if snake head is on food, grow its size
	if(next_tile == FOOD) {
		++snake_size;
		if(snake_size == NLEDS)
			return 2;
		play_food_sound();
		spawn_food();
	} else if(next_tile == SNAKE){ // you lost
		play_death_sound();
		return 1;
	} else {
		game_area[snake[tail].x][snake[tail].y] = EMPTY;
		tail = SAFE_INC(tail, NLEDS);
	}
	
	game_area[snake[head].x][snake[head].y] = SNAKE;
	return 0;
}


void init_io(void) {
	// select buttons as input
	DDRB &= ~(1 << PB0);
	PORTB |= (1 << PB0);
	
	DDRB &= ~(1 << PB1);
	PORTB |= (1 << PB1);

	DDRB &= ~(1 << PB2);
	PORTB |= (1 << PB2);
	
	DDRB &= ~(1 << PB3);
	PORTB |= (1 << PB3);
	
	DDRB &= ~(1 << PB4);
	PORTB |= (1 << PB4);
	
	// select buzzer as output
	DDRB |= (1 << PB6);
	
	// activate interruptions
	sei();

	// activate buttons on interruptions
	PCICR  = (1 << PCIE1);
    PCMSK1 = (1 << PCINT8) | (1 << PCINT9) | (1 << PCINT10) | (1 << PCINT11) | (1 << PCINT12);
}

int main(void)
{
	//init ledmatrix
	init();
	setintensity(MX, 8);
	init_io();
	
	print_scrolling_message("Welcome to LEDSnake. Press start. ");
	start:
	print_scrolling_message("Select speed ");
	speed = select_speed();
	
	init_snake();
	spawn_food();
	print_to_matrix(game_area);
	
	// game loop
	while(1)
	{
		print_to_matrix(game_area);
		int ret = move();
		direction_changed = 0;
		if(ret == 1) {
			print_scrolling_message("You lost :(  ");
			goto start;
		}
		if(ret == 2) {
			play_sound(1300, 50);
			print_scrolling_message("You won :)  ");
			goto start;
		}
		_delay_ms(speed);
	}
	
	return 0;
}